
import datetime
import os
import time
import math
import warnings

from sentinel5dl import search, download
import cartopy.crs as ccrs
import cartopy.feature as cfeature
import matplotlib.pyplot as plt
import matplotlib.patheffects as PathEffects
import numpy as np
import pandas as pd
import xarray as xr

import ddeq

# define dictionary with all sentinel 5P products
sentinel_products = {
    "UV Aerosol Index": "L2__AER_AI",
    "Aerosol Layer Height": "L2__AER_LH",
    "Methane": "L2__CH4___",
    "Cloud": "L2__CLOUD_",
    "Carbon Monoxide": "L2__CO____",
    "Formaldehyde": "L2__HCHO__",
    "Nitrogen Dioxide": "L2__NO2___",
    "Ozone": "L2__O3____",
    "Ozone Profiles": "L2__O3_PR_",
    "Tropospheric Ozone": "L2__O3_TCL",
    "Sulphur Dioxide": "L2__SO2___",
}

# define dictionary of important variables for each product
product_variables = {
    "L2__CH4___": {
        "keep_variables": [
            "time_utc",
            "qa_value",
            "methane_mixing_ratio",
            "methane_mixing_ratio_precision",
            "methane_mixing_ratio_bias_corrected"
        ]
    },
    "L2__CLOUD_": {
        "keep_variables": [
            "time_utc",
            "qa_value",
            "cloud_fraction",
            "cloud_fraction_precision",
            "cloud_top_pressure",
            "cloud_top_pressure_precision",
            "cloud_base_pressure",
            "cloud_base_pressure_precision",
            "cloud_top_height",
            "cloud_top_height_precision",
            "cloud_base_height",
            "cloud_base_height_precision",
            "cloud_optical_thickness",
            "cloud_optical_thickness_precision"
        ]
    },
    "L2__CO____": {
        "keep_variables": [
            "time_utc",
            "qa_value",
            "carbonmonoxide_total_column",
            "carbonmonoxide_total_column_precision",
            "carbonmonoxide_total_column_corrected"
        ]
    },
    "L2__HCHO__": {
        "keep_variables": [
            "time_utc",
            "qa_value",
            "formaldehyde_tropospheric_vertical_column",
            "formaldehyde_tropospheric_vertical_column_precision"
        ]
    },
    "L2__NO2___": {
        "keep_variables": [
            "time_utc",
            "qa_value",
            "nitrogendioxide_tropospheric_column",
            "nitrogendioxide_tropospheric_column_precision",
            "averaging_kernel",
            "air_mass_factor_troposphere"
        ]
    },
    "L2__O3____":  {
        "keep_variables": [
            "time_utc",
            "qa_value",
            "ozone_total_vertical_column",
            "ozone_total_vertical_column_precision"
        ]
    },
    "L2__SO2___": {
        "keep_variables": [
            "time_utc",
            "qa_value",
            "sulfurdioxide_total_vertical_column",
            "sulfurdioxide_total_vertical_column_precision"
        ]
    }
}


def source_coordinates(sources):
    """
    Extract coordinates for the given sources from the library of plumes and
    saves them to a dictionary.

    Paramaters
    ----------
    sources : list
        Sources for which the coordinates have to be extracted.

    Returns
    -------
    coord_dict : dict
        Dictionary of the specified sources with a tuple of their longitudes and
        latitudes.
    """
    coord_dict = {}
    all_sources = ddeq.misc.read_point_sources()

    for source in sources:
        coords = (
            float(all_sources.lon_o.sel(source=source).values),
            float(all_sources.lat_o.sel(source=source).values)
        )
        coord_dict[source] = coords

    return coord_dict


def get_bounding_box(source, distance, coord_dict):
    """
    Calculates the minimum and maximum longitudes and latitudes of a box around
    a given source.

    Paramaters
    ----------
    source : str
        Source around which the box should be calculated.
    distance : float
        Distance between source and box to be calculated in km.
    coord_dict : dict
        Dictonary of all available sources and their coordinates.

    Returns
    -------
    (min_lat, min_lon, max_lat, max_lon) : float
        Tuple of minimum and maximum longitudes and latitudes.
    """
    lon, lat = coord_dict[source]

    R = 6378.137 # Radius of earth in KM
    lat1 = lat * math.pi / 180 # convert latitude to radians
    lon1 = lon * math.pi / 180 # convert longitude to radians
    d = distance / R # convert distance to radians

    # Calculate the corners of the bounding box
    min_lat = lat1 - d
    max_lat = lat1 + d
    min_lon = lon1 - d / math.cos(lat1)
    max_lon = lon1 + d / math.cos(lat1)

    # Convert back to degrees
    min_lat = min_lat * 180 / math.pi
    max_lat = max_lat * 180 / math.pi
    min_lon = min_lon * 180 / math.pi
    max_lon = max_lon * 180 / math.pi

    return (min_lon, max_lon, min_lat, max_lat)


def area_of_interest(min_lon, max_lon, min_lat, max_lat):
    """
    Format a string with the given minimum and maximum longitudes and latitudes
    to create the bounding polygon around a source.
    """
    polygon = (
        "POLYGON(("
        f"{min_lon} {min_lat},"  # Bottom-Left
        f"{min_lon} {max_lat},"  # Top-Left
        f"{max_lon} {max_lat},"  # Top-Right
        f"{max_lon} {min_lat},"  # Bottom-Right
        f"{min_lon} {min_lat}))" # Bottom-Left
    )
    return polygon


def select_processing_mode(start_date, end_date):
    """
    Select the processing mode ('Offline' / 'Reprocessing') based on the start
    and end date for the NO2 product. From 2022-07-25 onwards, data version
    02.04 is available in the 'Offline' product. Before that, the reprocessed
    data is available in the 'Reprocessing' product. It is recommended to use
    the RPRO products in the orbit range 24655 - 24779, for which also OFFL
    products are available. Orbit 24779 has the start date
    2022-07-25T22:13:50.000Z.

    For more information see: https://sentinel.esa.int/documents/247904/3541451
        /Sentinel-5P-Nitrogen-Dioxide-Level-2-Product-Readme-File

    However: Reprocessing only available until orbit 22060 (2022-01-15).
    Return a string of the processing mode.
    """
    date_fmt = "%Y-%m-%dT%H:%M:%S.%fZ"
    date_threshold = datetime.datetime.strptime(
        "2022-07-25T22:13:50.000Z", "%Y-%m-%dT%H:%M:%S.%fZ"
    )

    if datetime.datetime.strptime(end_date, date_fmt) <= date_threshold:
        processing_mode = "Reprocessing"
    elif datetime.datetime.strptime(start_date, date_fmt) > date_threshold:
        processing_mode = "Offline"
    else:
        warnings.warn(
            f"Select a period which does not include {date_threshold}"
        )

    return processing_mode


def search_orbits(sources, distance, start_date, end_date, product,
                  processing_mode, processing_level="L2", per_request_limit=100,
                  print_results=True):
    """
    Search orbits based on sources, start and end date, product and processing
    level. Store the files found to download them.

    Paramaters
    ----------
    sources : list
        Sources for which the coordinates have to be extracted.
    distance : float
        Distance between source and box to be calculated in km.
    start_date : str
        Start date of sensing period, in ISO 8601 format,
        i.e "2022-07-25T22:13:50.000Z"
    end_date : str
        End date of sensing period, in ISO 8601 format,
        i.e "2022-07-25T22:13:50.000Z"
    product : str
        Sentinel 5P product to analyse, ("UV Aerosol Index",
        "Aerosol Layer Height", "Methane", "Cloud", "Carbon Monoxide",
        "Formaldehyde", "Nitrogen Dioxide", "Ozone", "Ozone Profiles",
        "Tropospheric Ozone", "Sulphur Dioxide")
    processing mode : str
        Data processing mode ('Offline','Near real time' or 'Reprocessing')
    processing level : str
        Processing level of Sentinel 5P product, "L1B" or "L2".
    per_request_limit : int
        Limit number of results per request.
    print_results : bool
        Indicate if file names of results should be printed.

    Returns
    -------
    all_files : list
        List of all JSON files found. Is needed as input for
        'sentinel5dl.download'
    all_filenames: list
        List of all file names found.
    """
    # save file names in list
    all_files = []
    all_filenames = []
    coord_dict = source_coordinates(sources)

    # TO DO: select newest version for other trace gases as well
    if product == "Nitrogen Dioxide":
        # select data version v02.04 only
        processing_mode = select_processing_mode(start_date, end_date)
    else:
        processing_mode = processing_mode

    for source in sources:
        result = search(
            polygon=area_of_interest(*get_bounding_box(source, distance,
                                                       coord_dict)),
            begin_ts=start_date,
            end_ts=end_date,
            product=sentinel_products[product],
            processing_level=processing_level,
            processing_mode=processing_mode,
            per_request_limit=100)

        print(f'{result["totalresults"]} file(s) found for {source}')

        for i in range(0,result["totalresults"]):
            if print_results:
                print(result["products"][i]["identifier"])
            if result["products"][i] not in all_files:
                all_files.append(result["products"][i])
                all_filenames.append(
                    f'{result["products"][i]["identifier"]}.nc'
                )
        print("\t")

    return all_files, all_filenames


def open_dataset(filename, path="./"):
    """
    Read in the data files from the given path. Return netCDF file.
    """
    try:
        data_S5p = xr.open_dataset(f"{path}{filename}", group="PRODUCT")

        # add original file name as attribute
        data_S5p.attrs["original file name"] = filename

        return data_S5p

    except FileNotFoundError:
        raise FileNotFoundError("There is no file called '{}' in '{}'".format(
            filename, path))


def reduce_dims_and_vars(data_S5p):
    """
    Remove the dimensions and variables from a netCDF file, which are not
    needed. Documentation of all variables: https://sentinel.esa.int/documents
    /247904/2474726/Sentinel-5P-Level-2-Product-User-Manual-Nitrogen-Dioxide.pdf

    Paramaters
    ----------
    data_S5p : netCDF
        File to be cleaned.

    Returns
    -------
    data_S5p : netCDF
        Cleaned file.
    """
    # get product name from dictionary.
    product = data_S5p.attrs["original file name"][9:19]

    # For some other products, lat and lon are variables instead of coordinates
    # => add to coordinates
    if product in ["L2__O3____", "L2__SO2___", "L2__HCHO__", "L2__CLOUD_"]:
        data_S5p = data_S5p.set_coords(("latitude", "longitude"))

    # get important dimensions and variables for each product
    keep_dimensions = ["scanline", "ground_pixel", "time", "latitude",
                       "longitude"]
    keep_variables = product_variables[product]["keep_variables"]

    # drop unneccessary dimensions and variables
    for dim in data_S5p.indexes:
        if dim not in keep_dimensions:
            data_S5p=data_S5p.drop_dims(dim)

    for var in data_S5p.data_vars:
        if var not in keep_variables:
            data_S5p=data_S5p.drop_vars(var)

    if len(data_S5p.dims) != 2:
        data_S5p = data_S5p.squeeze("time")

    return data_S5p


def crop_data(data_S5p, source, distance, coord_dict):
    """
    Crop the data to the latitudes and longitudes which were used to select and
    download the data. Add attributes for the source to which the data has been
    cropped and the distance around the source in the resulting netCDF file.

    Paramaters
    ----------
    data : netCDF
        File to be cropped.
    source : str
        Source to which the data should be cropped.
    distance : float
        Distance between source and box to be calculated in km.
    coord_dict : dict
        Dictonary of all available sources and their coordinates.

    Returns
    -------
    data_S5p_cropped : netCDF
        Cropped file.
    save : bool
        Indicate if the dataset is not empty and should be saved.
    """
    # get min and max lon and lat around the source
    min_lon, max_lon, min_lat, max_lat = get_bounding_box(source, distance,
                                                          coord_dict)
    # crop data to area needed
    mask_lon = (data_S5p.longitude >= min_lon) & (data_S5p.longitude <= max_lon)
    mask_lat = (data_S5p.latitude >= min_lat) & (data_S5p.latitude <= max_lat)

    # check if there is data in the area of interest. Use variable which is in
    # all products
    if np.count_nonzero(~np.isnan(np.where(
        mask_lon & mask_lat, data_S5p.qa_value, np.nan))
                       ) > 0:
        data_S5p_cropped = data_S5p.where(mask_lon & mask_lat, drop=True)
        save = True
    else:
        data_S5p_cropped = data_S5p.where(mask_lon & mask_lat)
        warnings.warn(f'The file {data_S5p.attrs["original file name"]} does '
                      f'not contain any data around {source}.')
        save = False

    # add attributes
    data_S5p_cropped.attrs["description"] = (
        "Sentinel-5P data - cropped to a specific source"
    )
    data_S5p_cropped.attrs["source"] = source
    data_S5p_cropped.attrs["distance around source [km]"] = distance

    return data_S5p_cropped, save


def save_file(data_S5p, path_open="./", path_save="./", delete=False):
    """
    Save a file to the indicated directory.
    If the directory does not exist, it is created.
    If the file already exists, the user is asked to discard changes or
    overwrite the file.

    Paramaters
    ----------
    data_S5p : netCDF
        File to be saved to netCDF.
    path_open : str
        Path where the original file is stored.
    path_save : str
        Path to where the cropped file should be stored.
    delete : bool
        Indicate if the input file should be deleted.
    """
    filename_output = (
        f'{data_S5p.attrs["source"]}_{data_S5p.attrs["original file name"]}'
    )
    file_exists = os.path.exists(f"{path_save}{filename_output}")

    if not os.path.isdir(path_save):
        os.makedirs(path_save)

    if file_exists:
        overwrite = input(
            f"[WARNING] {filename_output} already exists in {path_save} "
            f"- overwrite? [y/n]"
        )

        while overwrite not in ["y", "n"]:
            overwrite = input("Enter 'y' (overwrite) or 'n' (cancel).")

        if overwrite == "n":
            print("File not saved")

        elif overwrite == "y":
            os.remove(f"{path_save}{filename_output}")
            # sleep for 5 seconds because overwriting does not work otherwise
            time.sleep(10)
            data_S5p.to_netcdf(path=f"{path_save}{filename_output}")
            print(f"{filename_output} replaced in directory {path_save}")
    else:
        data_S5p.to_netcdf(path=f"{path_save}{filename_output}")
        print(f"{filename_output} saved in directory {path_save}")

    if delete:
        os.remove(f'{path_open}{data_S5p.attrs["original file name"]}')


def crop_and_save(all_filenames, sources, distance, path_open="./",
                  path_save="./"):
    """
    Crop and save data.

    Paramaters
    ----------
    all_filenames : netCDF
        File to be saved to netCDF.
    sources : list
        Sources for which the data has to be cropped and saved. 
    distance : float
        Distance between source and box to be calculated in km.
    path_open : str
        Path where the original file is stored.
    path_save : str
        Path to where the cropped file should be stored.
    """
    coord_dict = source_coordinates(sources)

    for file in all_filenames:
        data_S5p = open_dataset(file, path=path_open)
        data_S5p = reduce_dims_and_vars(data_S5p)

        for source in sources:
            data_S5p_cropped, save = crop_data(data_S5p, source, distance,
                                               coord_dict)
            if save:
                save_file(data_S5p_cropped, path_open=path_open,
                          path_save=path_save, delete=False)


def plot_extent(data_S5p, var, coord_dict, vmin=None, vmax=None, zoom=True,
                qa=True, ha="left", va="center"):
    """
    Create a map of the variable indicated.

    Paramaters
    ----------
    data_S5p : netCDF
        File from which the data should be plotted.
    var : str
        Variable which should be plotted
    coord_dict : dict
        Dictionary of all available sources and their coordinates.
    vmin : float
        Minimum value on the colourbar.
    vmax : float
        Maximum value on the colourbar.
    zoom : bool
        Indicate if the map should be zoomed to the minimum and maximum
        latitudes and longitudes which were used to select and download the
        data.
    qa : bool
        Only show data with a qa_value > 0.75.
    ha : str
        Horizontal alignment of the text ["left", "center", "right"]
    va : str
        Vertical alignment of the text ["bottom", "center", "top"]

    Returns
    -------
    Plot the data.
    """

    lon_o, lat_o = coord_dict[data_S5p.attrs["source"]]
    min_lon, max_lon, min_lat, max_lat = get_bounding_box(
        data_S5p.attrs["source"], data_S5p.attrs["distance around source [km]"],
        coord_dict)
    lon = data_S5p.longitude
    lat = data_S5p.latitude
    value = data_S5p[var]
    label = data_S5p[var].long_name
    units = data_S5p[var].units

    # filter values with qa_value > 0.75
    if qa:
        value = np.where(data_S5p["qa_value"] > 0.75, value, np.nan)

    if vmin == None and vmax == None:
        vmin = np.nanquantile(value, 0.05)
        vmax = np.nanquantile(value, 0.95)

    # calculate offset for text so that it does not overlap with the point
    # marker
    if ha == "left":
        x_offset = 0.15
    elif ha == "right":
        x_offset = -0.15

    # plot map and gridlines
    plt.figure(figsize=(10,5))
    ax = plt.axes(projection=ccrs.PlateCarree())
    ax.coastlines(resolution="10m", color="k", linewidth=1.0)
    ax.axis("equal")
    lines = cfeature.NaturalEarthFeature(
        category="cultural",
        name="admin_0_boundary_lines_land",
        scale="10m")
    ax.add_feature(lines, edgecolor="k", facecolor="none", linewidth=1.0)
    gl = ax.gridlines(crs=ccrs.PlateCarree(), draw_labels=True)
    gl.top_labels = False
    gl.right_labels = False

    # plot data and source
    m = ax.pcolormesh(lon, lat, value, vmin=vmin, vmax=vmax, cmap="viridis",
                      transform=ccrs.PlateCarree(), shading="auto")
    ax.scatter(lon_o, lat_o, marker="o", s=20, c="black", edgecolor="white",
               transform=ccrs.PlateCarree())
    ax.text(lon_o+x_offset, lat_o, data_S5p.attrs["source"], clip_on=True,
            horizontalalignment=ha,
            verticalalignment=va,
            path_effects=[PathEffects.withStroke(linewidth=2.5,
                                                 foreground="w")])
    # \n to prevent the label from being cut off
    plt.colorbar(m).set_label(label=f"{label} [{units}] \n", wrap=True)
    plt.title(pd.to_datetime(data_S5p.time.values).strftime("%Y-%m-%d"))

    if zoom:
        ax.set_xlim(min_lon, max_lon)
        ax.set_ylim(min_lat, max_lat)

    plt.show()


def plot_orbit(file, product, variable="nitrogendioxide_tropospheric_column",
               path="./"):
    """
    Plot the whole orbit from the downloaded netCDF file.
    """
    # read in data and remove dimensions/variables
    data_S5p = open_dataset(file, path)
    data_S5p = reduce_dims_and_vars(data_S5p)

    vmin = np.nanquantile(data_S5p[variable], 0.05)
    vmax = np.nanquantile(data_S5p[variable], 0.95)
    orbit = file[52:57] # extract orbit number from filename

    ax = plt.axes(projection=ccrs.PlateCarree())
    ax.set_global(); ax.coastlines()
    data_S5p[variable].plot.pcolormesh(ax=ax, x='longitude', y='latitude',
                                       add_colorbar=True, cmap='viridis',
                                       vmin=vmin, vmax=vmax)

    plt.title(
        f'time = {pd.to_datetime(data_S5p.time.values).strftime("%Y-%m-%d")}, '
        f' orbit = {orbit}'
    )

    plt.show()