
import calendar
import os
import warnings

import cartopy.crs as ccrs
import cdsapi
import numpy as np
import pandas
import xarray as xr

import ddeq

from ddeq.smartcarb import DOMAIN


def read_smartcarb(time, lon, lat, radius=None, data_path='.', method='linear',
                   average=False):
    """\
    Read SMARTCARB winds at given location at given `time`. The location is
    interpolated from SMARTCARB model grid to given `lon` and `lat`. If `lon`
    and `lat` are given as scalar, it is possible to provide a radius around
    location for which are extracted.

    radius :: radius around location given in rotated degrees
    data_path :: path to SMARTCARB wind fields
    method :: interpolation method (used by xr.DataArray.interp method)
    average :: average extracted winds

    Return xr.Dataset with wind components `U` and `V` as well as `wind speed`
    and `direction`.
    """
    lon = np.atleast_1d(lon).flatten()
    lat = np.atleast_1d(lat).flatten()

    if lon.size > 1 and radius is not None:
        raise ValueError(
            'Please provide lon and lat as scalar if radius is given.'
        )

    rlon, rlat, _ = np.transpose(
        DOMAIN.proj.transform_points(ccrs.PlateCarree(), lon, lat)
    )

    filename = time.round('H').strftime('SMARTCARB_winds_%Y%m%d%H.nc')
    filename = os.path.join(data_path, filename)

    with xr.open_dataset(filename) as nc:
        if radius is None:
            u = [float(nc['U_GNFR_A'].interp(rlon=rlon[i], rlat=rlat[i],
                                             method=method))
                 for i in range(rlon.size)]

            v = [float(nc['V_GNFR_A'].interp(rlon=rlon[i], rlat=rlat[i],
                                             method=method))
                 for i in range(rlon.size)]

        else:
            distance = np.sqrt((nc.rlon - rlon)**2 + (nc.rlat - rlat)**2)
            region = nc.where(distance <= radius, drop=True)

            lon = region.lon.values.flatten()
            lat = region.lat.values.flatten()
            u = region['U_GNFR_A'].values.flatten()
            v = region['V_GNFR_A'].values.flatten()

    wind = xr.Dataset()
    wind['lon'] = xr.DataArray(lon, dims='index')
    wind['lat'] = xr.DataArray(lat, dims='index')
    wind['U'] = xr.DataArray(u, dims='index')
    wind['V'] = xr.DataArray(v, dims='index')

    if average:
        u_std = wind['U'].std()
        v_std = wind['V'].std()
        wind = wind.mean('index')
        wind['U_std'] = u_std
        wind['V_std'] = v_std

    wind['speed'] = np.sqrt(wind.U**2 + wind.V**2)
    wind['direction'] = ddeq.wind.calculate_wind_direction(wind.U, wind.V)

    return wind.where(np.isfinite(wind['U']), drop=True)


def read_era5(time, lon_o, lat_o, suffix=None):
    """
    Read ERA-5 data from CoCO2 project on ICOS-CP.
    """
    warnings.warn('CoCO2-specific function will be removed in future.',
                  DeprecationWarning, stacklevel=2)

    if suffix is None:
        filename = time.strftime(
            '/project/coco2/jupyter/WP4/ERA5/ERA5-gnfra-%Y%m%dt%H00.nc'
        )
    else:
        filename = time.strftime(
            f'/project/coco2/jupyter/WP4/ERA5/ERA5-gnfra-%Y%m%dt%H00_{suffix}.nc'
        )

    with xr.open_dataset(filename) as nc:

        u = nc['U_GNFR_A'].sel(longitude=lon_o, latitude=lat_o, method='nearest')
        v = nc['V_GNFR_A'].sel(longitude=lon_o, latitude=lat_o, method='nearest')

        U = float(np.sqrt(u**2 + v**2))
        D = ddeq.wind.calculate_wind_direction(float(u), float(v))

    return U, D


def calculate_wind_direction(u, v):
    return (270.0 - np.rad2deg(np.arctan2(v,u))) % 360


def get_wind(time, lon, lat, data_path, components='uv'):
    
    p, u, v = get_era5_wind_profile(time, lon, lat, data_path=data_path)
    
    # TODO: use weights
    u = u.mean('level')
    v = v.mean('level')
        
    if components == 'uv':
        return u, v
    
    else:
        wind_speed = np.sqrt(u**2 + v**2)
        wind_direction = calculate_wind_direction(u, v)
        return wind_speed, wind_direction
    

    
def download_era5_winds(time, data_path='.', overwrite=False, full_day=False):
   
    if full_day:
        era5_filename = os.path.join(data_path, time.strftime('ERA5-wind-%Y%m%d.nc'))
    else:
        era5_filename = os.path.join(data_path, time.strftime('ERA5-wind-%Y%m%d_%H%M.nc'))
    
    if os.path.exists(era5_filename) and not overwrite:
        return
    
    cds = cdsapi.Client()
    
    query = {
        'product_type': 'reanalysis',
        'format': 'netcdf',
        'variable': [
            'geopotential', 'temperature', 'u_component_of_wind',
            'v_component_of_wind',
        ],
        'pressure_level': [
            '700', '750', '775',
            '800', '825', '850',
            '875', '900', '925',
            '950', '975', '1000',
        ],
        "area": [90, -180, -90, 180], # north, east, south, west
    }
    
    if full_day:
        query['year'] = f'{time.year}'
        query['month'] = f'{time.month:02d}'
        query['day'] = f'{time.day:02d}'
        query['time'] = [
                '00:00', '01:00', '02:00',
                '03:00', '04:00', '05:00',
                '06:00', '07:00', '08:00',
                '09:00', '10:00', '11:00',
                '12:00', '13:00', '14:00',
                '15:00', '16:00', '17:00',
                '18:00', '19:00', '20:00',
                '21:00', '22:00', '23:00',
        ]
    else:
        query["date"] = time.strftime('%Y-%m-%d')
        query["time"] = time.strftime('%H:00')

    r = cds.retrieve('reanalysis-era5-pressure-levels', query, era5_filename)
    

def get_era5_wind_profile(time, lon, lat, data_path='.'):

    era5_filename = os.path.join(data_path, time.strftime('ERA5-wind-%Y%m%d_%H%M.nc'))
    download_era5_winds(time, data_path, overwrite=False)

    with xr.open_dataset(era5_filename) as era5:
        era5 = era5.load().sel(time=time, longitude=lon, latitude=lat, method='nearest')

    pressure = era5['level']
    u_profile = era5['u']
    v_profile = era5['v']  
    
    return pressure, u_profile, v_profile



def calculate_pressure_levels(surface_pressure, nlevels=37, which='middle'):
    
    filename= os.path.join(os.path.dirname(__file__), 'data',
                           'ERA5 L137 model level definitions.csv')
    
    level_definitions = pandas.read_csv(filename, index_col=0)

    a = np.array(level_definitions['a [Pa]'])
    b = np.array(level_definitions['b'])
    
    # pressure at boundary of model layers
    ph = a + b * float(surface_pressure)
   
    # TODO: how to interpolate?
    #ph = np.interp(np.linspace(0,138,nlevels+1), np.arange(138), ph)

    # pressure at middle of model layers
    pf = 0.5 * (ph[1:] + ph[:-1])
    
    if which == 'middle':
        return pf
    elif which == 'boundary':
        return ph
    else:
        raise ValueError


def calculate_level_heights(temperature, pressure, which='middle'):
    """
    temperature at middle of layers
    pressure at boundary of layers
    """
    g = 9.80665
    R = 287.06
  
    
    p1 = pressure[1:]
    p2 = pressure[:-1]

    dh = R * temperature / g * np.log(p1 / p2)
    
    hh = np.concatenate([[0.0], dh[::-1].cumsum()])
    hh = hh[::-1]
    
    hf = 0.5 * (hh[1:] + hh[:-1])
    
    if which == 'middle':
        return hf
    elif which == 'boundary':
        return hh
    else:
        raise ValueError
    
    
